from typing import Dict, Optional

from requests_html import HTML, HTMLSession

from nitter_scraper.schema import Profile  # noqa: I100, I202


def username_cleaner(username: str) -> str:
    """Strips @ symbol from a username.

    Example:
        @dgnsrekt -> dgnsrekt

    Args:
        username: username with @ symbol to remove.

    Returns:
        Username with @ symbol stripped.
    """
    return username.replace("@", "")


def link_parser(element: HTML) -> str:
    """Gets the first link from an html element

    Used for the profiles website, photo and banner links.

    Args:
        element: HTML element with a link to parse.

    Returns:
        First link from a collection of links.
    """
    return list(element.links)[0]


def parse_user_id_from_banner(banner_url: str) -> str:
    """Parses the users id from the users banner photo url.

    The user id can only be parsed from the banner photos url.

    Example:
    ```
        /pic/profile_banners%2F2474416796%2F1600567028%2F1500x500 -> 2474416796
                               ^        ^
                               |        |
                               ----------
                               user id section in banner link
    ```

    Args:
        banner_url: URL of the profiles banner photo.

    Returns:
        The target profiles user id.

    """
    return banner_url.split("%2F")[1]


def stat_cleaner(stat: str) -> int:
    """Cleans and converts single stat.

    Used for the tweets, followers, following, and likes count sections.

    Args:
        stat: Stat to be cleaned.

    Returns:
        A stat with commas removed and converted to int.

    """
    return int(stat.replace(",", ""))


def profile_parser(elements: Dict) -> Dict:
    """Converts parsed sections to text.

    Cleans and processes a dictionary of gathered html elements.

    Args:
        elements: Elements prepared to clean and convert.

    Returns:
        A dictionary of element sections cleaned and converted to their finalized types.

    """
    elements["username"] = username_cleaner(elements["username"].text)

    elements["name"] = elements["name"].text

    if elements.get("location"):
        elements["location"] = elements["location"].text

    elements["is_verified"] = True if elements.get("is_verified") else False

    if elements.get("biography"):
        elements["biography"] = elements["biography"].text

    if elements.get("website"):
        elements["website"] = link_parser(elements["website"])

    if elements.get("profile_photo"):
        elements["profile_photo"] = link_parser(elements["profile_photo"])

    if elements.get("banner_photo"):
        elements["banner_photo"] = link_parser(elements["banner_photo"])
        elements["user_id"] = parse_user_id_from_banner(elements["banner_photo"])

    if elements.get("tweets_count"):
        elements["tweets_count"] = stat_cleaner(elements["tweets_count"].text)

    if elements.get("following_count"):
        elements["following_count"] = stat_cleaner(elements["following_count"].text)

    if elements.get("followers_count"):
        elements["followers_count"] = stat_cleaner(elements["followers_count"].text)

    if elements.get("likes_count"):
        elements["likes_count"] = stat_cleaner(elements["likes_count"].text)

    return elements


def html_parser(html: HTML) -> Dict:
    """Parses HTML element into individual sections

    Given an html element the html_parser will search for each profile section using
    CSS selectors. All parsed html elements are gathered into a dictionary and returned.

    Args:
        html: HTML element from a successful nitter profile scraped response.

    Returns:
        A dictionary of found elements from the parsed sections.

    """
    elements = {}

    elements["username"] = html.find(".profile-card-username", first=True)

    elements["name"] = html.find(".profile-card-fullname", first=True)

    elements["biography"] = html.find(".profile-bio", first=True)

    elements["location"] = html.find(".profile-location", first=True)

    elements["is_verified"] = html.find(
        ".profile-card-fullname .icon-container .verified-icon", first=True
    )

    elements["profile_photo"] = html.find(".profile-card-avatar", first=True)

    elements["banner_photo"] = html.find(".profile-banner a", first=True)

    elements["website"] = html.find(".profile-website", first=True)

    profile_statlist = html.find(".profile-statlist", first=True)

    elements["tweets_count"] = profile_statlist.find(".posts .profile-stat-num", first=True)

    elements["following_count"] = profile_statlist.find(".following .profile-stat-num", first=True)

    elements["followers_count"] = profile_statlist.find(".followers .profile-stat-num", first=True)

    elements["likes_count"] = profile_statlist.find(".likes .profile-stat-num", first=True)

    elements = {k: v for k, v in elements.items() if v is not None}

    return elements


def get_profile(
    username: str, not_found_ok: bool = False, address: str = "https://nitter.net"
) -> Optional[Profile]:
    """Scrapes nitter for the target users profile information.

    Args:
        username: The target profiles username.
        not_found_ok: If not_found_ok is false (the default), a ValueError is raised if the target
            profile doesn't exist. If not_found_ok is true, None will be returned instead.
        address: The address to scrape profile data from. The default scrape location is
            'https://nitter.net' which should be used as a backup. This value will normally be
            replaced by the address of a local docker container instance of nitter.

    Returns:
        Profile object if successfully scraped, otherwise None.

    Raises:
        ValueError: If the target profile does not exist and the not_found_ok argument is false.


    """
    url = f"{address}/{username}"
    session = HTMLSession()
    response = session.get(url)

    if response.status_code == 200:  # user exists
        elements = html_parser(response.html)
        parsed_elements = profile_parser(elements)
        return Profile.from_dict(parsed_elements)

    if not_found_ok:
        return None

    else:
        raise ValueError(f'Oops! Either "{username}" does not exist or is private.')



from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import BaseModel as Base


class Entries(Base):
    """Contains metadata parsed from the text contents of the tweet.

    This object is a subclass of the pydantic BaseModel which makes it easy to serialize
    the object with the .dict() and json() methods.

    Attributes:
        hashtags: Contains all hashtags parsed from the tweet. Example #Bitcoin.
        cashtags: Contains all cashtags parsed from the tweet. Example $BTC
        urls: Contains all URLs parsed from the tweet excluding photo/video media links.
        photos: Contains all URLs that link to photo media.
        videos: Contains all URLs that link to video media.
    """

    hashtags: List[str]
    cashtags: List[str]
    urls: List[str]
    photos: List[str]
    videos: List[str]


class Tweet(Base):
    """Represents a status update from a twitter user.

    This object is a subclass of the pydantic BaseModel which makes it easy to serialize
    the object with the .dict() and json() methods.

    Attributes:
        tweet_id: Twitter assigned id associated with the tweet.
        tweet_url: Twitter assigned url that links to the tweet.
        is_retweet: Represents if the tweet is a retweet.
        is_pinned: Represents if the user has pinned the tweet.
        time: Time the user sent the tweet.
        text: Text contents of the tweet.
        replies: A count of the replies to the tweet.
        retweets: A count of the times the tweet was retweeted.
        likes: A count of the times the tweet was liked.
        entries: Contains the entries object which holds metadata
            on the tweets text contents.

    """

    tweet_id: int
    tweet_url: str
    username: str
    # user_id: int # unable to parse from nitter.
    is_retweet: bool
    is_pinned: bool
    time: datetime
    text: str
    replies: int
    retweets: int
    likes: int
    entries: Entries

    @classmethod
    def from_dict(cls, elements: Dict[str, Any]) -> "Tweet":
        """Creates Tweet object from a dictionary of processed text elements.

        Args:
            elements: Preprocessed attributes of a tweet object.

        Returns:
            Tweet object.

        """
        return cls(**elements)


class Profile(Base):
    """The profile object contains Twitter User account metadata.

    This object is a subclass of the pydantic BaseModel which makes it easy to serialize
    the object with the .dict() and json() methods.

    Attributes:
        username: The users screen_name, handle or alias. '@dgnsrekt'
        name: The users name as they've defined it.
        profile_photo: URL reference to the profiles photo.
        tweets_count: The number of Tweets (including retweets) issued by the user.
        following_count: The number of accounts the user is following.
        followers_count: The number of followers this account has.
        likes_count: Number of likes the follower has received.
        is_verified: Indicates if the user has been verified.
        banner_photo: URL reference to the profiles banner.
        biography: The users autobiography.
        user_id: User identification number generated by twitter.
        location: A user defined location.
        website: A user defined website.

    """

    username: str
    name: str
    # birthday: str #NOTE: need to find a twitter page with birthday to parse.
    profile_photo: str
    tweets_count: int
    following_count: int
    followers_count: int
    likes_count: int
    is_verified: bool
    # is_private: bool

    banner_photo: Optional[str] = None
    biography: Optional[str] = None
    user_id: Optional[int] = None
    location: Optional[str] = None
    website: Optional[str] = None

    @classmethod
    def from_dict(cls, elements: Dict[str, Any]) -> "Profile":
        """Creates Profile object from a dictionary of processed text elements.

        Args:
            elements: Preprocessed attributes of a profile object.

        Returns:
            Profile object.

        """
        return cls(**elements)



from pathlib import Path

SOURCE_ROOT = Path(__file__).parent
"""* A path to the nitter_scraper source code directory."""

PROJECT_ROOT = SOURCE_ROOT.parent
"""* A path to the nitter_scraper project directory."""

TEMPLATES_DIRECTORY = SOURCE_ROOT / "templates"
"""* A path to the template directory. The template is used by
jinja2 to render docker config files."""

TEST_DIRECTORY = PROJECT_ROOT / "tests"
"""* A path to the nitter_scraper test directory"""



from contextlib import contextmanager
from ipaddress import IPv4Address
from tempfile import _TemporaryFileWrapper as TemporaryFile
from tempfile import NamedTemporaryFile
import time
from typing import ClassVar, Optional

import docker
from docker.client import DockerClient
from docker.models.containers import Container
from jinja2 import Environment, FileSystemLoader
from loguru import logger
from pydantic import BaseModel as Base

from nitter_scraper.paths import PROJECT_ROOT, TEMPLATES_DIRECTORY  # noqa: I202, I100
from nitter_scraper.profile import get_profile  # noqa: I202, I100
from nitter_scraper.tweets import get_tweets  # noqa: I202, I100


class DockerBase(Base):
    """Provides helper methods for connecting to the docker client."""

    client: ClassVar[DockerClient] = None

    @classmethod
    def _get_client(cls):
        if cls.client is None:
            cls.client = docker.from_env()

        cls.client.ping()
        logger.info("Docker connection successful.")

        return cls.client


class Nitter(DockerBase):
    """Nitter Docker container object

    Args:
        host (IPv4Address): The host address the docker container will bind too.
        port (int): The port the docker container will listen to.

    Attributes:
        tempfile (TemporaryFile): A TemporaryFile file generated from a template.
        container (Container): Local representation of a container object.
            Holds the started instance of a docker container.
        address (str): The full address of the docker container.
        ports (dict[int, int]): Binds the listening port to the nitter docker container's
            internal port 8080.
        config_filepath (str): Path name to the generated tempfile.
        volumes (dict[str, dict[str, str]]): used to configure a bind volume.


    """

    host: IPv4Address
    port: int

    tempfile: TemporaryFile = None
    container: Optional[Container]

    class Config:
        arbitrary_types_allowed = True

    @property
    def address(self):
        return f"http://{self.host}:{self.port}"

    @property
    def config_filepath(self):
        if self.tempfile:
            return self.tempfile.name

    @property
    def ports(self):
        return {8080: self.port}

    @property
    def volumes(self):
        volumes = {"bind": "/src/nitter.conf", "mode": "ro"}
        return {self.config_filepath: volumes}

    def _render_config(self):
        env = Environment(loader=FileSystemLoader(TEMPLATES_DIRECTORY))
        template = env.get_template("nitter.conf")
        return template.render(self.dict())

    def _create_configfile(self):
        config = self._render_config()
        self.tempfile = NamedTemporaryFile(dir=PROJECT_ROOT)
        self.tempfile.write(config.encode())
        self.tempfile.seek(0)

    def get_profile(self, username: str, not_found_ok: bool = False):
        """Scrapes nitter for the target users profile information.

        This is a modified version of nitter_scraper.profile.get_profile().
        This version automatically uses the address of the docker container as the primary
        address to scrape profile data.

        Args:
            username: The target profiles username.
            not_found_ok: If not_found_ok is false (the default), a ValueError is raised if
                the target profile doesn't exist. If not_found_ok is true, None will be returned
                instead.

        Returns:
            Profile object if successfully scraped, otherwise None.

        Raises:
            ValueError: If the target profile does not exist and the not_found_ok argument is
                false.
        """
        return get_profile(username=username, not_found_ok=not_found_ok, address=self.address)

    def get_tweets(self, username: str, pages: int = 25, break_on_tweet_id: Optional[int] = None):
        """Gets the target users tweets

        This is a modified version of nitter_scraper.tweets.get_tweets().
        This version automatically uses the address of the docker container as the primary
        address to scrape profile data.

        Args:
            username: Targeted users username.
            pages: Max number of pages to lookback starting from the latest tweet.
            break_on_tweet_id: Gives the ability to break out of a loop if a tweets id is found.
            address: The address to scrape from. The default is https://nitter.net which should
                be used as a fallback address.

        Yields:
            Tweet Objects

        """

        return get_tweets(
            username=username,
            pages=pages,
            break_on_tweet_id=break_on_tweet_id,
            address=self.address,
        )

    def start(self):
        """Starts the docker the container"""
        self._create_configfile()
        client = self._get_client()

        self.container = client.containers.run(
            image="zedeus/nitter:2c6cabb4abe79166ce9973d8652fb213c1b0c5a2",
            auto_remove=True,
            ports=self.ports,
            detach=True,
            volumes=self.volumes,
        )
        time.sleep(1)
        logger.info(f"Running container {self.container.name} {self.container.short_id}.")

    def stop(self):
        """Stops the docker the container"""
        logger.info(f"Stopping container {self.container.name} {self.container.short_id}.")
        if self.container:
            self.container.stop(timeout=5)
            logger.info(f"Container {self.container.name} {self.container.short_id} Destroyed.")


@contextmanager
def NitterScraper(host: str = "0.0.0.0", port: int = 8080):
    """The NitterScraper context manager.

    Takes care of configuring, starting, and stopping a docker instance of nitter.

    Args:
        host: The host address the docker container will bind too.
        port: The port the docker container will listen to.

    Yields:
        Nitter: An object representing a started nitter docker container.
    """
    nitter = Nitter(host=host, port=port)
    nitter.start()

    try:
        yield nitter

    finally:
        nitter.stop()



from nitter_scraper.nitter import NitterScraper
from nitter_scraper.profile import get_profile
from nitter_scraper.tweets import get_tweets

__all__ = ["get_profile", "get_tweets", "NitterScraper"]

__version__ = "0.5.0"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nitter_scraper']

package_data = \
{'': ['*'], 'nitter_scraper': ['templates/*']}

install_requires = \
['docker>=4.3.1,<5.0.0',
 'jinja2>=2.11.2,<3.0.0',
 'loguru>=0.5.1,<0.6.0',
 'pendulum>=2.1.2,<3.0.0',
 'pydantic>=1.6.1,<2.0.0',
 'requests-html>=0.10.0,<0.11.0']

setup_kwargs = {
    'name': 'nitter-scraper',
    'version': '0.5.0',
    'description': 'Scrape Twitter API without authentication using Nitter.',
    'long_description': '# Nitter Scraper\n\nNitter Scraper is for anyone who enjoys the [twitter-scraper](https://github.com/bisguzar/twitter-scraper/) library. Nitter Scraper leverages running a local docker container instance of [nitter](https://github.com/zedeus/nitter) to scrape a users tweets and profile information without the twitter api ratelimit. This api works similar to the [twitter-scraper](https://github.com/bisguzar/twitter-scraper/) project with a few differences.\n\n## Docker Engine\nFor the best experience use this library with [Docker Engine](https://docs.docker.com/engine/) properly installed. The NitterScraper manager will start, stop and remove a docker instance of [nitter](https://hub.docker.com/r/zedeus/nitter). If you can\'t run docker you can import the get_tweets and get_profile functions to scrape from [nitter.net](http://www.nitter.net).\n\n\n## Getting Started\n\n### Prereqs\n* Docker Engine\n* Python ^3.7\n\n### Install\n```shell\npip install nitter-scraper\n```\n\n#### How to Scrape a twitter users profile information.\n```python\nfrom pprint import pprint\n\nfrom nitter_scraper import NitterScraper\n\nwith NitterScraper(host="0.0.0.0", port=8008) as nitter:\n    profile = nitter.get_profile("dgnsrekt")\n    print("serialize to json\\n")\n    print(profile.json(indent=4))\n    print("serialize to a dictionary\\n")\n    pprint(profile.dict())\n\n```\n#### Output\n```python\n$ python3 examples/basic_usage.py\n2020-09-21 18:11:23.429 | INFO     | nitter_scraper.nitter:_get_client:31 - Docker connection successful.\n2020-09-21 18:11:25.102 | INFO     | nitter_scraper.nitter:start:135 - Running container infallible_noyce 91122c9b7b.\nserialize to json\n\n{\n    "username": "DGNSREKT",\n    "name": "DGNSREKT",\n    "profile_photo": "/pic/profile_images%2F1307990704384245760%2FSBVd3XT6.png",\n    "tweets_count": 2897,\n    "following_count": 904,\n    "followers_count": 117,\n    "likes_count": 4992,\n    "is_verified": false,\n    "banner_photo": "/pic/profile_banners%2F2474416796%2F1600684261%2F1500x500",\n    "biography": "BITCOIN IS DEAD AGAIN. :(",\n    "user_id": 2474416796,\n    "location": "Moon",\n    "website": "https://github.com/dgnsrekt"\n}\nserialize to a dictionary\n\n{\'banner_photo\': \'/pic/profile_banners%2F2474416796%2F1600684261%2F1500x500\',\n \'biography\': \'BITCOIN IS DEAD AGAIN. :(\',\n \'followers_count\': 117,\n \'following_count\': 904,\n \'is_verified\': False,\n \'likes_count\': 4992,\n \'location\': \'Moon\',\n \'name\': \'DGNSREKT\',\n \'profile_photo\': \'/pic/profile_images%2F1307990704384245760%2FSBVd3XT6.png\',\n \'tweets_count\': 2897,\n \'user_id\': 2474416796,\n \'username\': \'DGNSREKT\',\n \'website\': \'https://github.com/dgnsrekt\'}\n2020-09-21 18:11:25.905 | INFO     | nitter_scraper.nitter:stop:139 - Stopping container infallible_noyce 91122c9b7b.\n2020-09-21 18:11:31.284 | INFO     | nitter_scraper.nitter:stop:142 - Container infallible_noyce 91122c9b7b Destroyed.\n\n\n```\n#### Next step run the [examples](https://nitter-scraper.readthedocs.io/en/latest/examples/)\n\n## NitterScraper Limitation\n\n* About max 800 tweets per user.\n* Unable to scrape trends from nitter.\n* To scrape the user_id the user must have a banner photo. If the banner photo url isn\'t present\nthe user_id will be none.\n* The user_id cannot be scraped from the tweets.\n* birthday and is_private are not implemented in the profile.\n\n## Contact Information\nTelegram = Twitter = Tradingview = Discord = @dgnsrekt\n\nEmail = dgnsrekt@pm.me\n',
    'author': 'dgnsrekt',
    'author_email': 'dgnsrekt@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://nitter-scraper.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Module for scraping tweets"""
from datetime import datetime
import re
from typing import Dict, Optional

from requests_html import HTMLSession

from nitter_scraper.schema import Tweet  # noqa: I100, I202


def link_parser(tweet_link):
    links = list(tweet_link.links)
    tweet_url = links[0]
    parts = links[0].split("/")

    tweet_id = parts[-1].replace("#m", "")
    username = parts[1]
    return tweet_id, username, tweet_url


def date_parser(tweet_date):
    split_datetime = tweet_date.split(",")

    day, month, year = split_datetime[0].strip().split("/")
    hour, minute, second = split_datetime[1].strip().split(":")

    data = {}

    data["day"] = int(day)
    data["month"] = int(month)
    data["year"] = int(year)

    data["hour"] = int(hour)
    data["minute"] = int(minute)
    data["second"] = int(second)

    return datetime(**data)


def clean_stat(stat):
    return int(stat.replace(",", ""))


def stats_parser(tweet_stats):
    stats = {}
    for ic in tweet_stats.find(".icon-container"):
        key = ic.find("span", first=True).attrs["class"][0].replace("icon", "").replace("-", "")
        value = ic.text
        stats[key] = value
    return stats


def attachment_parser(attachements):
    photos, videos = [], []
    if attachements:
        photos = [i.attrs["src"] for i in attachements.find("img")]
        videos = [i.attrs["src"] for i in attachements.find("source")]
    return photos, videos


def cashtag_parser(text):
    cashtag_regex = re.compile(r"\$[^\d\s]\w*")
    return cashtag_regex.findall(text)


def hashtag_parser(text):
    hashtag_regex = re.compile(r"\#[^\d\s]\w*")
    return hashtag_regex.findall(text)


def url_parser(links):
    return sorted(filter(lambda link: "http://" in link or "https://" in link, links))


def parse_tweet(html) -> Dict:
    data = {}
    id, username, url = link_parser(html.find(".tweet-link", first=True))
    data["tweet_id"] = id
    data["tweet_url"] = url
    data["username"] = username

    retweet = html.find(".retweet-header .icon-container .icon-retweet", first=True)
    data["is_retweet"] = True if retweet else False

    body = html.find(".tweet-body", first=True)

    pinned = body.find(".pinned", first=True)
    data["is_pinned"] = True if pinned is not None else False

    data["time"] = date_parser(body.find(".tweet-date a", first=True).attrs["title"])

    content = body.find(".tweet-content", first=True)
    data["text"] = content.text

    # tweet_header = html.find(".tweet-header") #NOTE: Maybe useful later on

    stats = stats_parser(html.find(".tweet-stats", first=True))

    if stats.get("comment"):
        data["replies"] = clean_stat(stats.get("comment"))

    if stats.get("retweet"):
        data["retweets"] = clean_stat(stats.get("retweet"))

    if stats.get("heart"):
        data["likes"] = clean_stat(stats.get("heart"))

    entries = {}
    entries["hashtags"] = hashtag_parser(content.text)
    entries["cashtags"] = cashtag_parser(content.text)
    entries["urls"] = url_parser(content.links)

    photos, videos = attachment_parser(body.find(".attachments", first=True))
    entries["photos"] = photos
    entries["videos"] = videos

    data["entries"] = entries
    # quote = html.find(".quote", first=True) #NOTE: Maybe useful later on
    return data


def timeline_parser(html):
    return html.find(".timeline", first=True)


def pagination_parser(timeline, address, username) -> str:
    next_page = list(timeline.find(".show-more")[-1].links)[0]
    return f"{address}/{username}{next_page}"


def get_tweets(
    username: str,
    pages: int = 25,
    break_on_tweet_id: Optional[int] = None,
    address="https://nitter.net",
) -> Tweet:
    """Gets the target users tweets

    Args:
        username: Targeted users username.
        pages: Max number of pages to lookback starting from the latest tweet.
        break_on_tweet_id: Gives the ability to break out of a loop if a tweets id is found.
        address: The address to scrape from. The default is https://nitter.net which should
            be used as a fallback address.

    Yields:
        Tweet Objects

    """
    url = f"{address}/{username}"
    session = HTMLSession()

    def gen_tweets(pages):
        response = session.get(url)

        while pages > 0:
            if response.status_code == 200:
                timeline = timeline_parser(response.html)

                next_url = pagination_parser(timeline, address, username)

                timeline_items = timeline.find(".timeline-item")

                for item in timeline_items:
                    if "show-more" in item.attrs["class"]:
                        continue

                    tweet_data = parse_tweet(item)
                    tweet = Tweet.from_dict(tweet_data)

                    if tweet.tweet_id == break_on_tweet_id:
                        pages = 0
                        break

                    yield tweet

            response = session.get(next_url)
            pages -= 1

    yield from gen_tweets(pages)
